# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
"""
Module containing LennardJonesFit class for curve fitting with Lennard-Jones.
"""

import random
import numpy as np  # pylint: disable=W0611
import matplotlib
from .fit_base import FitBase, log

matplotlib.use("Agg")


class LennardJonesFit(FitBase):
    """
    This class makes curve fitting using Lennard-Jones potential formula.
    It reads data from a file given as the first argument.
    The data file must contain at least two columns.
    The first column represents x values,
    the second and following columns are the y values.
    """

    def __init__(self, filename, **kwargs):
        """
        **Arguments:**

        filename:
            (string) Name of the file that contains data.

        **Keyword arguments:**

        guess:
            (tuple) Guess of the optimal parameters for the model.
        shift
            (bool) If true, shifts the data
        """

        self.common_init(filename, **kwargs)
        self.type_of_fit = "Lennard-Jones potential"
        self.label_type = "lennard_jones_fit"

    def model_function(self, x, *args, **kwargs):
        """Depends on the module."""
        r = x
        if len(args) != 2:
            raise TypeError("Wrong amount of guess parameters passed")
        sigma, epsilon = args
        return 4 * epsilon * ((sigma / r) ** 12 - (sigma / r) ** 6)

    def default_guess(self, context):
        """Default guess adequate for the model function."""
        guess = [1, 1]

        return guess

    def random_guess(self, context):
        """
        Random guess adequate for the model function with constraints
        chosen by the user.
        """
        if self.auto_range is None:

            guess = [random.uniform(-60, 60), random.uniform(-60, 60)]
        else:
            guess = [
                random.uniform(min_val, max_val)
                for min_val, max_val in self.auto_range
            ]

        return guess

    def print_info(self, popt):
        """
        Print info about popt in the format adequate for the model function.
        """
        name_temp = ("sigma", "epsilon")
        for el in range(2):
            log(f"\t {name_temp[el]}: {popt[el]:2.8f}")



"""
Module containing tests for LennardJonesFit class.
"""

# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#

import pytest
from .common import check_list_within_percentage
from pathlib import Path
from .. import lennard_jones

# pylint: skip-file


def test_LennardJonesFit():
    """Simple general test"""
    l_j_1 = lennard_jones.LennardJonesFit(
        "foo_data.txt", auto_guess=True, precision=5, shift=True
    )
    l_j_1()
    assert l_j_1.sq_root_error <= 10
    file_path = (
        Path.cwd() / "minifit-results" / "lennard_jones_fit_foo_data_2.pdf"
    )
    assert file_path.exists(), f"File {file_path} does not exist"


def test_popt_LennardJonesFit():
    l_j_1 = lennard_jones.LennardJonesFit(
        "foo_data.txt", auto_guess=False, guess=(0.75, 0.26), shift=True
    )
    l_j_1()
    assert check_list_within_percentage(
        l_j_1.popt, (0.75353688, 0.26085384), 10
    )
    assert (
        check_list_within_percentage(
            l_j_1.popt, (0.75353688 * 1.1, 0.26085384), 1
        )
        is not True
    )
    # first element differes by 10%, arrays should not be considered equal if precison is set to 1%


def test_LennardJonesFit_model_function():
    l_j_1 = lennard_jones.LennardJonesFit("exp.dat")
    with pytest.raises(TypeError):
        l_j_1.model_function(-1)
    with pytest.raises(TypeError):
        l_j_1.model_function(1, 2, 3, 4)
    assert l_j_1.model_function(0.01, 0.02, 0.039) == 628.992



# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
"""
Module containing PolyFit class for curve fitting with polynomials.
"""
import random
import matplotlib
import numpy as np  # pylint: disable=W0611
from .fit_base import FitBase, log


matplotlib.use("Agg")


class PolyFit(FitBase):
    """
    This class makes a polynomial curve fitting.
    Order of the polynomial can be chosen.
    It reads data from a file given as the first argument.
    The data file must contain at least two columns.
    The first column represents x values,
    the second and following columns are the y values.
    """

    def __init__(self, filename, **kwargs):
        """
        **Arguments:**

        filename:
            (string) Name of the file that contains data.

        **Keyword arguments:**

        guess:
            (tuple) Guess of the optimal parameters for the model.
        order:
            (int) Order of the polynomial. Optional. Default 1
        """

        if kwargs.get("order") is None:
            log("=")
            log("Order of the polynomial wasn't given. Using default (1)")
            self.order = 1
        else:
            self.order = kwargs.get("order")
        self.common_init(filename, **kwargs)
        self.type_of_fit = "Polynomials"
        self.label_type = "polynomial_fit"

    def model_function(self, x, *args, **kwargs):
        """Depends on the module."""
        if len(args) != self.order + 1:
            raise TypeError("Wrong amount of guess parameters passed")
        val = 0.0
        # a + b*x + c*x^2 + etc.
        for exp in range(self.order + 1):
            val += args[exp] * x**exp
        return val

    def default_guess(self, context):
        """Default guess adequate for the model function."""
        guess = []
        for _ in range(self.order + 1):
            guess.append(1.0)
        return guess

    def random_guess(self, context):
        """
        Random guess adequate for the model function with constraints
        chosen by the user.
        """
        if self.auto_range is None:
            guess = []
            for _ in range(self.order + 1):
                guess.append(random.uniform(-150, 150))

        else:
            guess = [
                random.uniform(min_val, max_val)
                for min_val, max_val in self.auto_range
            ]

        return guess

    def print_info(self, popt):
        """
        Print info about popt in the format adequate for the model function.
        """
        for el in range(self.order + 1):
            log(f"\t {el}: {self.popt[el]:2.8f}")



"""
Module containing tests for PolyFit class.
"""

# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#

# pylint: skip-file

import pytest
from .common import check_list_within_percentage
from pathlib import Path
from .. import poly


def test_PolyFit():
    """Simple general test"""
    poly_1 = poly.PolyFit("example.dat", order=7)
    poly_1()
    assert poly_1.sq_root_error <= 6
    file_path = Path.cwd() / "minifit-results" / "polynomial_fit_example_3.pdf"
    assert file_path.exists(), f"File {file_path} does not exist"


def test_popt_PolyFit():
    poly_1 = poly.PolyFit("example.dat", order=7)
    poly_1()
    assert check_list_within_percentage(
        poly_1.popt,
        (
            -39.0169376799999981,
            -305.9215619700000275,
            572.6581489500000544,
            -599.6565376300000025,
            379.0770024099999773,
            -144.1919395300000133,
            30.4520343900000015,
            -2.7466285300000002,
        ),
        1,
    )
    assert (
        check_list_within_percentage(
            poly_1.popt,
            (
                -39.01693768 * 1.1,
                -305.92156197,
                572.65814895,
                -599.65653763,
                379.07700241,
                -144.19193953,
                30.45203439,
                -2.74662853,
            ),
            1,
        )
        is not True
    )
    # first element differes by 10%, arrays should not be considered equal if precison is set to 1%


def test_PolyFit_model_function():
    """Tests for model function"""
    poly_1 = poly.PolyFit("exp.dat", order=7)
    with pytest.raises(TypeError):
        poly_1.model_function(-1)
    with pytest.raises(TypeError):
        poly_1.model_function(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
    assert (
        poly_1.model_function(-5, -1, -2, -3, 4, 5, 0.003, 0.03, 25000)
        == -1953121981.625
    )



"""
Module containing tests for GaussFit class.
"""

# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#

# pylint: skip-file

import pytest
from .common import check_list_within_percentage
from pathlib import Path
from .. import gauss


def test_GaussFit():
    """Simple general test"""
    gauss_1 = gauss.GaussFit("gauss_data.txt")
    gauss_1()
    assert gauss_1.sq_root_error <= 10
    file_path = Path.cwd() / "minifit-results" / "gauss_fit_gauss_data_1.pdf"
    assert file_path.exists(), f"File {file_path} does not exist"


def test_popt_GaussFit():
    gauss_1 = gauss.GaussFit("gauss_data.txt")
    gauss_1()
    assert check_list_within_percentage(
        gauss_1.popt, (4.97587424, 20.00233527, 2.84768795), 1
    )
    assert (
        check_list_within_percentage(
            gauss_1.popt, (4.97587424 * 1.1, 20.00233527, 2.84768795), 1
        )
        is not True
    )
    # first element differes by 10%, arrays should not be considered equal if precison is set to 1%


def test_GaussFit_model_function():
    gauss_1 = gauss.GaussFit("exp.dat")
    with pytest.raises(TypeError):
        gauss_1.model_function(-1)
    with pytest.raises(TypeError):
        gauss_1.model_function(1, 2, 3, 4, 5)
    assert gauss_1.model_function(22, 44, 66, 22) == 0.8058881111043039



"""
Module containing tests for ExponentialFit class.
"""

# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#

# pylint: skip-file


import pytest
from .common import check_list_within_percentage
from pathlib import Path
from .. import exp


def test_ExponentialFit():
    """Simple general test"""
    exp_1 = exp.ExponentialFit("exp.dat")
    exp_1()
    assert exp_1.sq_root_error <= 20
    file_path = Path.cwd() / "minifit-results" / "exp_fit_exp_1.pdf"
    assert file_path.exists(), f"File {file_path} does not exist"


def test_popt_ExponentialFit():
    exp_1 = exp.ExponentialFit("exp.dat")
    exp_1()
    assert check_list_within_percentage(
        exp_1.popt, (2.70136445, 0.48028958, 0.79216127), 1
    )
    assert (
        check_list_within_percentage(
            exp_1.popt, (2.70136445 * 1.1, 0.48028958, 0.79216127), 1
        )
        is not True
    )
    # first element differes by 10%, arrays should not be considered equal if precison is set to 1%


def test_ExponentialFit_model_function():
    exp_1 = exp.ExponentialFit("exp.dat")
    with pytest.raises(TypeError):
        exp_1.model_function(-1)
    with pytest.raises(TypeError):
        exp_1.model_function(1, 2, 3, 4, 5)
    assert exp_1.model_function(3, 4, 2, -1000) == 613.7151739709404



# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
"""
Module containing logger and FitBase class.
"""

from abc import ABC, abstractmethod
import warnings
from pathlib import Path
from scipy.optimize import curve_fit
import matplotlib.pyplot as pyplot
import numpy as np


def log(str1):
    """Logger"""
    if str1 == "~":
        print("~" * 100)
    elif str1 == "=":
        print("=" * 100)
    elif str1 == "-":
        print("-" * 100)
    elif str1 == " ":
        print(" ")
    else:
        print(f"\t{str1}")


# pylint: disable=unnecessary-pass,attribute-defined-outside-init


class FitBase(ABC):
    """
    Base class for all fitting classes
    """

    def common_init(self, filename, **kwargs):
        """
        Common constructor
        """
        self.type_of_fit = (
            "FitBase instance attribute - needs to be overwritten"
        )
        self.label_type = (
            "FitBase instance attribute - needs to be overwritten"
        )
        path_data = Path("./minifit-data/" + filename)
        path_cwd = Path.cwd()
        if path_data.exists() is not True and path_cwd.exists is not True:
            raise FileNotFoundError(
                f"\"Couldn't find {filename} in minifit-data"
            )

        self.filename = filename

        if kwargs.get("guess") is None:
            self.guess = None
        else:
            self.guess = kwargs.get("guess")

        # self.guess = kwargs.get("guess", None)

        if kwargs.get("auto_guess") is True:
            self.auto_guess = True
            self.auto_range = kwargs.get("auto_range", None)
        else:
            self.auto_guess = False
            self.auto_range = None

        if kwargs.get("precision") is not None:
            self.precision = kwargs.get("precision")
        else:
            self.precision = 0.4
        # self.precision = kwargs.get("precision",  0.4)

        if kwargs.get("shift") is None:
            self.shift = False
        else:
            self.shift = kwargs.get("shift")
        # self.precision = kwargs.get("shift",  False)

        if path_data.exists() is True:
            self.filename = filename
            self.read_data(str(path_data))
        else:
            self.filename = filename
            self.read_data(str(path_cwd))

        self.popt = None
        self.pcov = None
        self.abs_sum_diff = np.inf
        self.sq_root_error = np.inf

    def read_data(self, filename):
        """Reads data from each column then shifts the data."""
        cols = None
        with open(filename) as file:
            for line in file:
                if not line.startswith("#"):
                    data = [float(x) for x in line.split()]
                    if cols is not None:
                        cols = np.vstack((cols, data))
                    else:
                        cols = np.array(data)
        self.xdata = cols[:, 0]
        self.ydata = cols[:, 1:]

        self.shift_data = self.ydata[-1, :].copy()
        if self.shift is True:
            self.ydata -= self.ydata[-1, :]
        else:
            self.shift_data[self.shift_data != 0] = 0.0

    @abstractmethod
    def model_function(self, x, *args, **kwargs):
        """Depends on the module."""
        return None

    @abstractmethod
    def default_guess(self, context):
        """Default guess adequate for the model function."""

    @abstractmethod
    def random_guess(self, context):
        """
        Random guess adequate for the model function with constraints
        chosen by the user.
        """

    @abstractmethod
    def print_info(self, popt):
        """
        Print info about popt in the format adequate for the model function.
        """

    def context_setter(self, xdata, ydata):
        """
        If overwritten can return data useful for setting guess in random_gues and default_guess
        """
        return [xdata, ydata]

    def optimize(self, xdata, ydata, ind):
        """Finds the optimal parameters and corresponding errors."""

        guess = self.default_guess(self.context_setter(xdata, ydata))
        if self.guess is not None:
            guess = self.guess
        if self.auto_guess is False:
            self.popt, self.pcov = curve_fit(
                self.model_function, xdata, ydata, p0=guess
            )
        else:
            flag = True
            while flag:
                try:
                    self.abs_sum_diff = np.inf
                    self.sq_root_error = np.inf
                    guess = self.random_guess(
                        self.context_setter(xdata, ydata)
                    )

                    self.popt, self.pcov = curve_fit(
                        self.model_function, xdata, ydata, p0=guess
                    )
                    diff = self.model_function(xdata, *self.popt) - ydata
                    self.abs_sum_diff = np.sum(abs(diff))
                    self.sq_root_error = self.sq_root_error = np.sqrt(
                        np.dot(diff, diff)
                    )
                except RuntimeError:
                    log(
                        "Unable to fit with the current guess. Trying again with a new one..."
                    )
                finally:
                    if (
                        self.sq_root_error < self.precision
                        and self.abs_sum_diff < self.precision
                    ):
                        flag = False

        log("Guess:")
        log(guess)
        log("Optimization succesful")

        log("Optimized fitting parameters:")

        self.print_info(self.popt)

        log("Differences between fit and data points:")
        diff = self.model_function(xdata, *self.popt) - ydata

        log("\t Data  \t\tModel\t      Difference")
        for x, y, z in zip(
            ydata, self.model_function(xdata, *self.popt), diff
        ):
            log(
                f"\t {x+self.shift_data[ind]:6.8f} {y+self.shift_data[ind]:6.8f}  {z:e}"
            )

        log("Covariance matrix:")
        print(f"{self.pcov}")
        self.abs_sum_diff = np.sum(abs(diff))
        log(f"Absolute sum of differences:  {self.abs_sum_diff:2.3e}")
        self.sq_root_error = np.sqrt(np.dot(diff, diff))
        log(f"Square root error:            {self.sq_root_error:2.3e}")

    def show(self, xdata, ydata, ind):
        """Saves graph of the results as pdf."""
        trial_x = np.linspace(xdata[0], xdata[-1], 1000)
        model_y = (
            self.model_function(trial_x, *self.popt) + self.shift_data[ind]
        )

        pyplot.figure()
        pyplot.plot(
            xdata, ydata + self.shift_data[ind], label="Data", marker="o"
        )
        pyplot.plot(trial_x, model_y, "r-", ls="--", label=self.label_type)
        pyplot.legend()
        pyplot.show()

        base = Path(self.filename)
        name = base.stem
        outname = self.label_type + "_" + name + "_" + str(ind + 1) + ".pdf"
        save_results_to = "./minifit-results/"
        log(f"Saving results to directory: {save_results_to}")
        mf_results = str(Path.cwd()) + "/minifit-results"
        if Path(mf_results).exists() is not True:
            Path(mf_results).mkdir()
        pyplot.savefig(save_results_to + outname)

    def __call__(self):
        """Calls optimize() and then show() for chosen data."""
        warnings.filterwarnings("ignore")

        log("=")
        log("MiniFit module for curve fitting with the least squares method")
        log(f"Using: {self.type_of_fit}")

        log("=")
        log(" ")
        log(f"Reading data from input file {self.filename}")
        log(" ")
        log(f"Input file contains {self.ydata.shape[1]} different states.")
        log("~")
        for ind in range(self.ydata.shape[1]):
            log(f"Fitting curve for state {ind+1}:")
            log("-")
            try:
                self.optimize(self.xdata, self.ydata[:, ind], ind)
                self.show(self.xdata, self.ydata[:, ind], ind)
            except RuntimeError:
                log(f"Optimization failed for data set {ind+1}")

            log("Done.")
            log("~")



"""
Module containing tests for MorseFit class.
"""

# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#

# pylint: skip-file

import pytest
from .common import check_list_within_percentage
from pathlib import Path
from .. import morse


def test_MorseFit():
    """Simple general test"""
    morse_1 = morse.MorseFit(
        "foo_data.txt",
        auto_guess=True,
        precision=5,
        shift=True,
    )
    morse_1()
    assert morse_1.sq_root_error <= 6
    file_path = Path.cwd() / "minifit-results" / "morse_fit_foo_data_2.pdf"
    assert file_path.exists(), f"File {file_path} does not exist"


def test_popt_MorseFit():
    morse_1 = morse.MorseFit(
        "foo_data.txt",
        auto_guess=True,
        precision=0.1,
        shift=True,
    )
    morse_1()
    # auto_guess=True, popt will be slightly different each time
    # while staying in chosen precison range
    assert check_list_within_percentage(
        morse_1.popt,
        (
            1.09696084,
            0.43085574,
            2.7629781,
            0.01177289,
            0.00580579,
            -0.00493962,
            0.0020099,
            -0.00020443,
            1.233e-05,
        ),
        1,
    )
    assert (
        check_list_within_percentage(
            morse_1.popt,
            (
                1.09696084 * 1.1,
                0.43085574,
                2.7629781,
                0.01177289,
                0.00580579,
                -0.00493962,
                0.0020099,
                -0.00020443,
                1.233e-05,
            ),
            1,
        )
        is not True
    )
    # first element differes by 10%, arrays should not be considered equal if precison is set to 1%


def test_MorseFit_model_function():
    morse_1 = morse.MorseFit("exp.dat")
    with pytest.raises(TypeError):
        morse_1.model_function(-1)
    with pytest.raises(TypeError):
        morse_1.model_function(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)
    assert (
        morse_1.model_function(
            0.06,
            0.05,  # re
            0.1,  # de
            10.0,  # b0
            0.1,  # b1
            0.1,  # b2
            0.001,  # b3
            0.001,  # b4
            0.0001,  # b5
            0.000001,
        )
        == -0.007570818470665069
    )



"""
Module containing functions shared by tests
"""

# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>

# pylint: skip-file


def check_list_within_percentage(list1, list2, percentage):
    assert len(list1) == len(list2)

    for value1, value2 in zip(list1, list2):
        difference = abs(value1 - value2)
        max_difference = abs(percentage * value1 / 100)
        if difference > max_difference:
            return False

    return True



# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
"""
Module containing MorseFit class for curve fitting with morse.
"""

import random
import numpy as np
import matplotlib
from .fit_base import FitBase, log

matplotlib.use("Agg")


class MorseFit(FitBase):
    """
    This class makes curve fitting using morse potential formula.
    It reads data from a file given as the first argument.
    The data file must contain at least two columns.
    The first column represents x values,
    the second and following columns are the y values.
    """

    def __init__(self, filename, **kwargs):
        """
        **Arguments:**

        filename:
            (string) Name of the file that contains data.

        **Keyword arguments:**

        guess:
            (tuple) Guess of the optimal parameters for the model.
        auto_guess:
            (bool) If true, tries to fit until chosen precision is reached. Default false.
        range_auto:
            (tuple) Sets boundaries for each parameter of guess.
            If not set, default range is used.
        precision:
            Used by auto_guess. Default 0.4
        shift:
            (bool) If true, shifts the data. Default false.
        """
        self.common_init(filename, **kwargs)
        self.type_of_fit = "Morse"
        self.label_type = "morse_fit"

    def model_function(self, x, *args, **kwargs):
        """Depends on the module."""
        if len(args) != 9:
            raise TypeError("Wrong amount of guess parameters passed")
        re, de, b0, b1, b2, b3, b4, b5, b6 = args
        return (
            de
            * (
                1
                - np.exp(
                    -b0
                    * (x - re)
                    / re
                    * (
                        1
                        + b1 * (b0 * (x - re) / re)
                        + b2 * (b0 * (x - re) / re) ** 2
                        + b3 * (b0 * (x - re) / re) ** 3
                        + b4 * (b0 * (x - re) / re) ** 4
                        + b5 * (b0 * (x - re) / re) ** 5
                        + b6 * (b0 * (x - re) / re) ** 6
                    )
                )
            )
            ** 2
            - de
        )

    def default_guess(self, context):
        """Default guess adequate for the model function."""
        guess = [
            context[0],  # re
            context[1],  # de
            10.0,  # b0
            0.1,  # b1
            0.1,  # b2
            0.001,  # b3
            0.001,  # b4
            0.0001,  # b5
            0.000001,  # b6
        ]
        return guess

    def random_guess(self, context):
        """
        Random guess adequate for the model function with constraints
        chosen by the user.
        """
        if self.auto_range is None:

            guess = [
                context[0],  # re
                context[1],  # de
                random.uniform(-50, 50),  # b0
                random.uniform(-10, 10),  # b1
                random.uniform(-10, 10),  # b2
                random.uniform(-10, 10),  # b3
                random.uniform(-0.1, 0.1),  # b4
                random.uniform(-0.1, 0.1),  # b5
                random.uniform(-0.1, 0.1),  # b6
            ]
        else:
            guess = [
                random.uniform(min_val, max_val)
                for min_val, max_val in self.auto_range
            ]

        return guess

    def context_setter(self, xdata, ydata):
        # returns data useful for setting guess in random_gues and default_guess

        re = xdata[np.argmin(ydata)]
        de = np.abs(np.min(ydata))
        return [re, de]

    def print_info(self, popt):
        """
        Print info about popt in the format adequate for the model function.
        """
        name_temp = ("re", "de", "b0", "b1", "b2", "b3", "b4", "b5", "b6")
        for el in range(9):
            log(f"\t {name_temp[el]}: {popt[el]:2.8f}")



"""
Module containing tests for MorseFit class.
"""

# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#

# pylint: skip-file

import pytest
from .common import check_list_within_percentage
from pathlib import Path
from .. import user_defined


def user_poly(x, *args, **kwargs):
    a, b, c, d, e, f, g, h = args
    return (
        a * x**7
        + b * x**6
        + c * x**5
        + d * x**4
        + e * x**3
        + f * x**2
        + g * x
        + h
    )


a_range = (
    (-2.67143698 - 50, -2.67143698 + 50),
    (29.75271303 - 50, 29.75271303 + 50),
    (-141.51011815 - 50, -141.51011815 + 50),
    (373.56237563 - 50, 373.56237563 + 50),
    (-593.05923681 - 50, -593.05923681 + 50),
    (568.07666599 - 50, 568.07666599 + 50),
    (-304.20455678 - 50, -304.20455678 + 50),
    (-39.28501939 - 50, -39.28501939 + 50),
)


def test_UserFit():
    """Simple general test"""
    user_1 = user_defined.UserFit(
        "example.dat",
        model=user_poly,
        num_param=8,
        auto_guess=True,
        precision=0.01,
        auto_range=a_range,
    )
    user_1()
    assert user_1.sq_root_error <= 1
    file_path = Path.cwd() / "minifit-results" / "user_fit_example_1.pdf"
    assert file_path.exists(), f"File {file_path} does not exist"


def test_popt_UserFit():
    user_1 = user_defined.UserFit(
        "example.dat",
        model=user_poly,
        num_param=8,
        auto_guess=True,
        precision=0.01,
        auto_range=a_range,
    )
    user_1()
    # auto_guess=True, popt will be slightly different each time while staying in chosen precison range
    assert check_list_within_percentage(
        user_1.popt,
        (
            -2.74664485,
            30.45219915,
            -144.19263652,
            379.07860216,
            -599.65868604,
            572.65983488,
            -305.92227695,
            -39.01681137,
        ),
        1,
    )
    assert (
        check_list_within_percentage(
            user_1.popt,
            (
                -2.67143698 * 1.10,
                29.75271303,
                -141.51011815,
                373.56237563,
                -593.05923681,
                568.07666599,
                -304.20455678,
                -39.28501939,
            ),
            1,
        )
        is not True
    )
    # first element differes by 10%, arrays should not be considered equal if precison is set to 1%



# -*- coding: utf-8 -*-
# Copyright (C) 2023-- Michał Kopczyński
#
# This file is part of MiniFit.
#
# MiniFit is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# MiniFit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
"""
Module containing MorseFit class for curve fitting with morse.
"""


import random
import matplotlib
import numpy as np  # pylint: disable=W0611
from .fit_base import FitBase, log


matplotlib.use("Agg")


class UserFit(FitBase):
    """
    This class makes curve fitting using morse potential formula.
    It reads data from a file given as the first argument.
    The data file must contain at least two columns.
    The first column represents x values,
    the second and following columns are the y values.
    """

    def __init__(self, filename, **kwargs):
        """
        **Arguments:**

        filename:
            (string) Name of the file that contains data.

        **Keyword arguments:**

        num_param:
            (int) Number of parameters of user defined model function.
        model:
            (function/callable) An object that expects (x, *args, **kwargs)
            and returns y value for given arguments
        guess:
            (tuple) Guess of the optimal parameters for the model.
        auto_guess:
            (bool) If true, tries to fit until chosen precision is reached. Default false.
        range_auto:
            (tuple) Sets boundaries for each parameter of guess.
            If not set, default range is used.
        precision:
            Used by auto_guess. Default 0.4
        shift:
            (bool) If true, shifts the data. Default false.
        """
        self.common_init(filename, **kwargs)

        self.type_of_fit = kwargs.get("type_of_fit", "User defined function")
        self.label_type = kwargs.get("label_type", "user_fit")
        num_param = kwargs.get("num_param")
        if num_param is None:
            raise ValueError(
                "Missing required parameter: num_param. How many parameters your functions has? Required for sanitazation"  # pylint: disable=C0301
            )
        self.num_param = num_param

        user_def_func = kwargs.get("model")
        if user_def_func is not None and callable(user_def_func):
            # The object is a function
            self.user_def_func = user_def_func
        else:
            # The object is not a function or is None
            raise ValueError(
                "Missing required parameter: model. With what function do you want to fit?"
            )

    def model_function(self, x, *args, **kwargs):
        """Depends on the module."""
        if len(args) != self.num_param:
            raise TypeError("Wrong amount of guess parameters passed")
        return self.user_def_func(x, *args, **kwargs)

    def default_guess(self, context):
        """Default guess adequate for the model function."""
        guess = []
        for _ in range(self.num_param):
            guess.append(1.0)
        return guess

    def random_guess(self, context):
        """
        Random guess adequate for the model function with constraints
        chosen by the user.
        """
        if self.auto_range is None:
            guess = []
            for _ in range(self.num_param):
                guess.append(random.uniform(-150, 150))

        else:
            guess = [
                random.uniform(min_val, max_val)
                for min_val, max_val in self.auto_range
            ]

        return guess

    def print_info(self, popt):
        """
        Print info about popt in the format adequate for the model function.
        """
        for el in range(self.num_param):
            log(f"\t {el}: {popt[el]:2.8f}")
